/*
  xcolumns - columns clone
  Copyright (C) 2007-2012 Jaromir Capik <tavvva@email.cz>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <curses.h>

#define false 0
#define true 1

#define ESC_KEY KeyArray[9]
#define LALT_KEY KeyArray[64]
#define Q_KEY KeyArray[24]
#define LEFT_KEY KeyArray[100]
#define RIGHT_KEY KeyArray[102]
#define UP_KEY KeyArray[98]
#define DOWN_KEY KeyArray[104]
#define SPACE_KEY KeyArray[65]
#define S_KEY KeyArray[39]

#define APP_NAME "Columns"
#define APP_ICON NULL
#define APP_CLASS "GAME_TAVVVA"
#define PERCENT_SIZE 0.85
#define CELL_COLORS 6
#define COLORS 10+CELL_COLORS*5

#define COLUMNS 6
#define ROWS 13

#define LABEL1 "XCOLUMNS v0.1.1"
#define LABEL2 "S - Start a new game"
#define LABEL3 "Q - Quit"
#define LABEL4 "Left - Move left"
#define LABEL5 "Right - Move right"
#define LABEL6 "Up - Rotate up"
#define LABEL7 "Down - Rotate down"
#define LABEL8 "Space - Drop"



Display *d;
int s;
Window w;
XEvent e;
Colormap cm;
GC gc;
bool KeyArray[255];
int KACount;
XSizeHints sh;
XClassHint ch;
XTextProperty win_name, icon_name;
int WIN_WIDTH,WIN_HEIGHT,DISP_WIDTH,DISP_HEIGHT;
int CELL_SIZE;
int cnt;
Status rc;
XColor Color[COLORS];
bool Redraw,InGame,Change;
int Dropping;
int Pos,Act[3],New[3],Tmp;
unsigned long score,score_cnt;
char str[20];
char *strlist[2];

int Cell[(COLUMNS+4)*(ROWS+4)];

int GetCell(signed long x,signed long y)
{
  return Cell[(x+2)+(y+2)*(COLUMNS+4)];
}

int PutCell(signed long x,signed long y,int color)
{
  Cell[(y+2)*(COLUMNS+4)+(x+2)]=color;
}

void DrawCell(signed long x,signed long y,int color)
{
  if (color==0) {
    XSetForeground(d, gc, Color[0].pixel);
    XFillRectangle(d, w, gc, 40+x*CELL_SIZE, (y+0.5)*CELL_SIZE, CELL_SIZE, CELL_SIZE);
    XSetForeground(d, gc, Color[2].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE, (y+0.5)*CELL_SIZE, 40+(x+1)*CELL_SIZE-2, (y+0.5)*CELL_SIZE);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE, (y+0.5)*CELL_SIZE, 40+x*CELL_SIZE, (y+1.5)*CELL_SIZE-2);
  } else {
    XSetForeground(d, gc, Color[2].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE, (y+0.5)*CELL_SIZE, 40+(x+1)*CELL_SIZE-2, (y+0.5)*CELL_SIZE);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE, (y+0.5)*CELL_SIZE, 40+x*CELL_SIZE, (y+1.5)*CELL_SIZE-2);
    XSetForeground(d, gc, Color[5+5*color+2].pixel);
    XFillRectangle(d, w, gc, 40+x*CELL_SIZE+3, (y+0.5)*CELL_SIZE+3, CELL_SIZE-5, CELL_SIZE-5);
    XSetForeground(d, gc, Color[5+5*color+1].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+1, (y+1.5)*CELL_SIZE-1, 40+(x+1)*CELL_SIZE-1, (y+1.5)*CELL_SIZE-1);
    XDrawLine(d,w,gc, 40+(x+1)*CELL_SIZE-1, (y+0.5)*CELL_SIZE+1, 40+(x+1)*CELL_SIZE-1, (y+1.5)*CELL_SIZE-1);
    XSetForeground(d, gc, Color[5+5*color+2].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+2, (y+1.5)*CELL_SIZE-2, 40+(x+1)*CELL_SIZE-2, (y+1.5)*CELL_SIZE-2);
    XDrawLine(d,w,gc, 40+(x+1)*CELL_SIZE-2, (y+0.5)*CELL_SIZE+2, 40+(x+1)*CELL_SIZE-2, (y+1.5)*CELL_SIZE-2);
    XSetForeground(d, gc, Color[5+5*color+3].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+2, (y+0.5)*CELL_SIZE+2, 40+(x+1)*CELL_SIZE-2, (y+0.5)*CELL_SIZE+2);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+2, (y+0.5)*CELL_SIZE+2, 40+x*CELL_SIZE+2, (y+1.5)*CELL_SIZE-2);
    XSetForeground(d, gc, Color[5+5*color+4].pixel);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+1, (y+0.5)*CELL_SIZE+1, 40+(x+1)*CELL_SIZE-1, (y+0.5)*CELL_SIZE+1);
    XDrawLine(d,w,gc, 40+x*CELL_SIZE+1, (y+0.5)*CELL_SIZE+1, 40+x*CELL_SIZE+1, (y+1.5)*CELL_SIZE-1);
  }
}

void DrawMap(void)
{
  signed long x,y;
  for (y=0;y<ROWS;y++) {
    for (x=0;x<COLUMNS;x++) {
      DrawCell(x,y,GetCell(x,y));
    }
  }
}

void ClearMap(void)
{
  signed long x,y;
  for (y=-2;y<ROWS+2;y++) {       // borders
    for (x=-2;x<COLUMNS+2;x++) {
      PutCell(x,y,10000);
    }
  }
  for (y=0;y<ROWS;y++) {
    for (x=0;x<COLUMNS;x++) {     // space
      PutCell(x,y,0);
    }
  }
}

void Generate(int Arr[3])
{
  Arr[0]=rand()%CELL_COLORS+1;
  Arr[1]=rand()%CELL_COLORS+1;
  Arr[2]=rand()%CELL_COLORS+1;
}

void Drop(void)
{
  long Top;
  Top=0;
  while(GetCell(Pos,Top++)==0);
  Top--;Top--;

  if (Top>1) {
    PutCell(Pos,Top--,Act[2]);
    PutCell(Pos,Top--,Act[1]);
    PutCell(Pos,Top--,Act[0]);
    Act[2]=New[2];
    Act[1]=New[1];
    Act[0]=New[0];
    Generate(New);

    DrawMap();
    InGame=false;
    Dropping=1;
    score_cnt=0;
  }
}

void EliminateCell(signed long x,signed long y)
{
  int color;
  color=GetCell(x,y);
  if (color!=0) {
    if ((color==(GetCell(x+1,y)&0xff))&&(color==(GetCell(x+2,y)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x-1,y)&0xff))&&(color==(GetCell(x-2,y)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x,y+1)&0xff))&&(color==(GetCell(x,y+2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x,y-1)&0xff))&&(color==(GetCell(x,y-2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x+1,y+1)&0xff))&&(color==(GetCell(x+2,y+2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x-1,y-1)&0xff))&&(color==(GetCell(x-2,y-2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x-1,y+1)&0xff))&&(color==(GetCell(x-2,y+2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x+1,y-1)&0xff))&&(color==(GetCell(x+2,y-2)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x+1,y)&0xff))&&(color==(GetCell(x-1,y)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x,y+1)&0xff))&&(color==(GetCell(x,y-1)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x+1,y+1)&0xff))&&(color==(GetCell(x-1,y-1)&0xff))) PutCell(x,y,color+256);
    if ((color==(GetCell(x+1,y-1)&0xff))&&(color==(GetCell(x-1,y+1)&0xff))) PutCell(x,y,color+256);
  }
}


void Sift(void)
{
  signed long x,y;
  long Top;
  for (y=ROWS-1;y>=0;y--) {
    for (x=0;x<COLUMNS;x++) {
      Top=y+1;
      if (GetCell(x,y)!=0) {
        while(GetCell(x,Top++)==0);Top--;Top--;
        if (y!=Top) {
          PutCell(x,Top,GetCell(x,y));
          PutCell(x,y,0);
        }
      }
    }
  }
  Redraw=true;
  Dropping=1;
}

bool Lost(void)
{
  signed long x,Top;
  bool Result;
  Result=true;

  for (x=0;x<COLUMNS;x++) {
    Top=0;
    while(GetCell(x,Top++)==0);Top--;
    if (Top>2) Result=false;
  }
  return Result;
}

void Eliminate(void)
{
  signed long x,y;

  Change=false;
  for (y=0;y<ROWS;y++) {
    for (x=0;x<COLUMNS;x++) {
      EliminateCell(x,y);
    }
  }
  for (y=0;y<ROWS;y++) {
    for (x=0;x<COLUMNS;x++) {
      if (GetCell(x,y)>256) {
        PutCell(x,y,0);
        Change=true;
        score_cnt++;
      }
    }
  }
  if (Change) Dropping=2;
  else {
    if (score_cnt>2) score+= 1<<(score_cnt-3);
    InGame=true;
    Dropping=0;
    Redraw=true;
    if (Lost()) {
      Dropping=3;
      InGame=false;
    }
  }
}


int main(int argc,char **argv)
{
// 0 - black
  Color[0].red   = 0;
  Color[0].green = 0;
  Color[0].blue  = 0;

// 1 - white
  Color[1].red   = 65535;
  Color[1].green = 65535;
  Color[1].blue  = 65535;

// 2 - darkgray
  Color[2].red   = 10000;
  Color[2].green = 10000;
  Color[2].blue  = 10000;

// 10..14 - red set
  Color[10].red   = 20000;
  Color[10].green = 0;
  Color[10].blue  = 0;
  Color[11].red   = 40000;
  Color[11].green = 0;
  Color[11].blue  = 0;
  Color[12].red   = 60000;
  Color[12].green = 0;
  Color[12].blue  = 0;
  Color[13].red   = 65535;
  Color[13].green = 17465;
  Color[13].blue  = 17465;
  Color[14].red   = 65535;
  Color[14].green = 34465;
  Color[14].blue  = 34465;

// 15..19 - green set
  Color[15].red   = 0;
  Color[15].green = 20000;
  Color[15].blue  = 0;
  Color[16].red   = 0;
  Color[16].green = 30000;
  Color[16].blue  = 0;
  Color[17].red   = 0;
  Color[17].green = 40000;
  Color[17].blue  = 0;
  Color[18].red   = 0;
  Color[18].green = 50000;
  Color[18].blue  = 0;
  Color[19].red   = 34465;
  Color[19].green = 65535;
  Color[19].blue  = 34465;

// 20..24 - blue set
  Color[20].red   = 0;
  Color[20].green = 0;
  Color[20].blue  = 20000;
  Color[21].red   = 0;
  Color[21].green = 0;
  Color[21].blue  = 40000;
  Color[22].red   = 0;
  Color[22].green = 0;
  Color[22].blue  = 60000;
  Color[23].red   = 24465;
  Color[23].green = 24465;
  Color[23].blue  = 65535;
  Color[24].red   = 44465;
  Color[24].green = 44465;
  Color[24].blue  = 65535;

// 25..29 - gray set
  Color[25].red   = 10000;
  Color[25].green = 10000;
  Color[25].blue  = 10000;
  Color[26].red   = 20000;
  Color[26].green = 20000;
  Color[26].blue  = 20000;
  Color[27].red   = 26000;
  Color[27].green = 26000;
  Color[27].blue  = 26000;
  Color[28].red   = 35000;
  Color[28].green = 35000;
  Color[28].blue  = 35000;
  Color[29].red   = 50000;
  Color[29].green = 50000;
  Color[29].blue  = 50000;

// 30..34 - purple set
  Color[30].red   = 20000;
  Color[30].green = 0;
  Color[30].blue  = 20000;
  Color[31].red   = 30000;
  Color[31].green = 0;
  Color[31].blue  = 30000;
  Color[32].red   = 40000;
  Color[32].green = 0;
  Color[32].blue  = 40000;
  Color[33].red   = 60000;
  Color[33].green = 0;
  Color[33].blue  = 60000;
  Color[34].red   = 65535;
  Color[34].green = 34465;
  Color[34].blue  = 65535;

// 35..39 - brown set
  Color[35].red   = 15000;
  Color[35].green = 8000;
  Color[35].blue  = 0;
  Color[36].red   = 20000;
  Color[36].green = 10000;
  Color[36].blue  = 0;
  Color[37].red   = 30000;
  Color[37].green = 15000;
  Color[37].blue  = 0;
  Color[38].red   = 40000;
  Color[38].green = 20000;
  Color[38].blue  = 0;
  Color[39].red   = 60000;
  Color[39].green = 30000;
  Color[39].blue  = 10000;


  InGame=false;
  Dropping=0;
  srand(time(NULL));

  score=0;
  Redraw=true;

  // key array init
  for(KACount=0;KACount<255;KACount++) KeyArray[KACount]=false;

  d=XOpenDisplay(NULL);
  if(d==NULL) {
    printf("Cannot open display\n");
    exit(1);
  }
  s=DefaultScreen(d);
  gc=DefaultGC(d, s);

  DISP_WIDTH=DisplayWidth(d,s);
  DISP_HEIGHT=DisplayHeight(d,s);
  WIN_HEIGHT=DISP_HEIGHT*PERCENT_SIZE;
  WIN_WIDTH=WIN_HEIGHT;

  CELL_SIZE=WIN_HEIGHT/(ROWS+1);

  sh.flags=PPosition | PSize | PMinSize | PMaxSize;
  sh.x=(DISP_WIDTH-WIN_WIDTH)/2;
  sh.y=(DISP_HEIGHT-WIN_HEIGHT)/2;
  sh.width=WIN_WIDTH;
  sh.height=WIN_HEIGHT;
  sh.min_width=WIN_WIDTH;
  sh.min_height=WIN_HEIGHT;
  sh.max_width=WIN_WIDTH;
  sh.max_height=WIN_HEIGHT;

  ch.res_name=APP_NAME;
  ch.res_class=APP_CLASS;

  strlist[0]=APP_NAME;
  strlist[1]=APP_ICON;
  XStringListToTextProperty(strlist,1,&win_name);
  XStringListToTextProperty(strlist,1,&icon_name);

  // create window
  w=XCreateSimpleWindow(d, RootWindow(d, s), sh.x, sh.y, sh.width, sh.height, 1, BlackPixel(d, s), WhitePixel(d, s));
  XSetWMProperties(d,w,&win_name,&icon_name,NULL,0,&sh,NULL,&ch);

  cm = DefaultColormap(d, DefaultScreen(d));	

  // color allocation
  for (cnt=0;cnt<COLORS;cnt++) {
    rc = XAllocColor(d, cm, &Color[cnt]);
    if (rc == 0) {
      fprintf(stderr, "XAllocColor - failed to allocate colors.\n");
      exit(1);
    }
  }

  // events
  XSelectInput(d, w, ExposureMask | KeyPressMask | KeyReleaseMask);

  // show the window
  XMapWindow(d, w);

  ClearMap();

  while(1) {

    XNextEvent(d, &e);

    if (Dropping==1) {
      napms(50);
      Eliminate();
      XRaiseWindow(d,w);
      XClearArea(d,w,0,0,1,1,true);
    } else if (Dropping==2) {
      napms(50);
      Sift();
      XRaiseWindow(d,w);
      XClearArea(d,w,0,0,1,1,true);
    } else if (Dropping==3) {
      Redraw=true;
    } 
    if(e.type==KeyPress) {
      //printf("%d\n",e.xkey.keycode);
      if (e.xkey.keycode<255) KeyArray[e.xkey.keycode]=true;
    }
    if(e.type==KeyRelease) {
      if (e.xkey.keycode<255) KeyArray[e.xkey.keycode]=false;
      //printf("%d\n",e.xkey.keycode);
    }
    if(e.type==KeyPress) {
      if (Q_KEY) break;
      if (S_KEY) {
        score=0;
        ClearMap();
        Pos=2;
        Generate(Act);
        Generate(New);
        InGame=true;
        Redraw=true;
        Dropping=0;
      }
      if (InGame) {
        if (LEFT_KEY) {
          if (Pos>0) {
            Pos--;
            Redraw=true;
          }
        }
        if (RIGHT_KEY) {
          if (Pos<(COLUMNS-1)) {
            Pos++;
            Redraw=true;
          }
        }
        if (UP_KEY) {
          Tmp=Act[0];
          Act[0]=Act[1];
          Act[1]=Act[2];
          Act[2]=Tmp;
          Redraw=true;
        }
        if (DOWN_KEY) {
          Tmp=Act[2];
          Act[2]=Act[1];
          Act[1]=Act[0];
          Act[0]=Tmp;
          Redraw=true;
        }
        if (SPACE_KEY) {
          SPACE_KEY=false;
          Drop();
        }
      }
    }
    // draw or redraw the window
    if((e.type==Expose)||(Redraw)) {
      // podklad
      Redraw=false;
      XSetForeground(d, gc, Color[2].pixel);
      XFillRectangle(d, w, gc, 0, 0, 40, WIN_HEIGHT);
      XFillRectangle(d, w, gc, 40+COLUMNS*CELL_SIZE, 0, WIN_WIDTH-(40+COLUMNS*CELL_SIZE), WIN_HEIGHT);
      XFillRectangle(d, w, gc, 40, 0                   , COLUMNS*CELL_SIZE, 0.5*CELL_SIZE);
      XFillRectangle(d, w, gc, 40, (0.5+ROWS)*CELL_SIZE, COLUMNS*CELL_SIZE, WIN_HEIGHT-(0.5+ROWS)*CELL_SIZE);
      XSetForeground(d, gc, Color[1].pixel);
      XDrawString(d,w,gc,10,10,LABEL1, sizeof(LABEL1)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,20,LABEL2, sizeof(LABEL2)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,40,LABEL3, sizeof(LABEL3)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,60,LABEL4, sizeof(LABEL4)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,80,LABEL5, sizeof(LABEL5)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,100,LABEL6, sizeof(LABEL6)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,120,LABEL7, sizeof(LABEL7)-1);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,140,LABEL8, sizeof(LABEL8)-1);

      sprintf(str,"Score : %u",score);
      XDrawString(d,w,gc,WIN_WIDTH*2/3,220,str, strlen(str));

      DrawMap();
      if (InGame) {
        DrawCell(Pos,0,Act[0]);
        DrawCell(Pos,1,Act[1]);
        DrawCell(Pos,2,Act[2]);
        DrawCell(COLUMNS+1,0,New[0]);
        DrawCell(COLUMNS+1,1,New[1]);
        DrawCell(COLUMNS+1,2,New[2]);
      }
      if (Dropping==3) {
        XSetForeground(d, gc, Color[1].pixel);
        XDrawString(d,w,gc,WIN_WIDTH*2/3,200,"--==GAME OVER==--", 17);
      }
    }
  }
  // close connection to server
  XCloseDisplay(d);
  return 0;
}
